#include <aversive.h>

struct servo {
	uint8_t bit;
	uint16_t command;
};

static struct servo servo_table[] = {
	{
		.bit = 0,
		.command = 0,
	},
	{
		.bit = 1,
		.command = 512,
	},
	{
		.bit = 2,
		.command = 1023,
	},
};
#define NB_SERVO (sizeof(servo_table)/sizeof(*servo_table))

register uint8_t bypass asm("r2");
register uint8_t done asm("r3");
register uint8_t portval asm("r4");
register uint8_t rxidx asm("r5");

/*
 * SPI protocol:
 *
 * A command is stored on 2 bytes. The first one has its msb to 0, and the
 * second one to 1. The first received byte contains the command number, and the
 * msb of the servo value. The second byte contains the lsb of the servo value.
 *
 * Commands 0 to NB_SERVO are to set the value of servo.
 * Command 14 is to enable bypass mode.
 * Command 15 is to disable bypass mode.
 */
register union {
	uint8_t u8;
	struct {
		uint16_t zero:1;
		uint16_t cmd_num:4;
		uint16_t val_msb:3;
	};
} byte0 asm("r6");

register union {
	uint8_t u8;
	struct {
		uint8_t one:1;
		uint8_t val_lsb:7;
	};
} byte1 asm("r7");

SIGNAL(TIMER1_COMPA_vect)
{
	PORTC = portval;
	TIMSK1 &= ~_BV(OCIE1A);
	done = 1;
}

static void poll_spi(void)
{
	uint8_t c;

	/* reception complete ? */
	if (!(SPSR & (1<<SPIF)))
		return;

	c = SPDR;
	if ((rxidx == 0) && (c & 0x80)) {
		rxidx = 0;
		return; /* drop */
	}
	if ((rxidx == 1) && ((c & 0x80) == 0)) {
		rxidx = 0;
		return; /* drop */
	}

	if (rxidx == 0) {
		byte0.u8 = c;
	}
	else {
		uint16_t val;

		byte1.u8 = c;

		/* process command */

		if (byte0.cmd_num < NB_SERVO) {
			val = (uint16_t)byte0.val_msb << 7;
			val += byte1.val_lsb;
			servo_table[byte0.cmd_num].command = val;
		}
		else if (byte0.cmd_num == 14) {
			bypass = 1;
		}
		else if (byte0.cmd_num == 15) {
			bypass = 0;
		}
	}

	rxidx ^= 1;
}

static void load_timer_at(uint16_t t)
{
	OCR1A = t;
	TIMSK1 |= _BV(OCIE1A);
}

static void do_one_servo(struct servo *s)
{
	uint16_t t;

	/* set bit */
	done = 0;
	//portval = PORTC | (1 << s->bit);
	portval = (1 << s->bit);
	t = TCNT1;
	load_timer_at(t + 150);
	while (done == 0)
		poll_spi();

	/* reset bit */
	done = 0;
	portval = 0;
	//portval = PORTC & (~(1 << s->bit));
	load_timer_at(t + 150 + 8000 + s->command * 8);
	while (done == 0)
		poll_spi();
}

int main(void)
{
	uint8_t i;
	uint8_t t, diff;

	/* LED */
	DDRB = 0x20;

	/* servo outputs */
	DDRC = 0x7;

	/* start timer1 at clk/1 (8Mhz) */
	TCNT1 = 0;
	TCCR1B = _BV(CS10);

	/* start timer0 at clk/1024 (~8Khz) */
	TCNT0 = 0;
	TCCR0B = _BV(CS02) | _BV(CS00);

	/* enable spi (don't set unused MISO as output) */
	SPCR = _BV(SPE);

	sei();

	while (1) {
		t = TCNT0;
		for (i = 0; i < NB_SERVO; i++) {
			do_one_servo(&servo_table[i]);
		}
		/* wait 20 ms */
		while (1) {
			diff = TCNT0 - t;
			if (diff >= 160)
				break;
			poll_spi();
		}
		/* bypass mode */
		while (bypass == 1) {
			PORTC = PORTB;
		}
	}

	return 0;
}
